from __future__ import annotations

import io
import math
import stat
import struct
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from unblob.file_utils import (
    Endian,
    FileSystem,
    InvalidInputFormat,
    StructParser,
)
from unblob.models import (
    Extractor,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


@dataclass
class MinixInode:
    i_mode: int
    i_nlinks: int
    i_uid: int
    i_gid: int
    i_size: int
    i_time: int | None
    i_atime: int | None
    i_mtime: int | None
    i_ctime: int | None
    i_zone: list[int]


@dataclass
class MinixDirEntry:
    inode: int
    name: bytes


# see linux/minix_fs.h
SUPERBLOCK_OFFSET = 0x400
ROOT_INODE_INDEX = 1
STATIC_BLOCK_SIZE = 1_024
HEADER_STRUCT = "minix_super_block"
INODE_V1_CDEF = """
    typedef struct minix_inode {
        uint16 i_mode;
        uint16 i_uid;
        uint32 i_size;
        uint32 i_time;
        uint8  i_gid;
        uint8  i_nlinks;
        uint16 i_zone[9];
    } minix_inode;
"""

INODE_V2_CDEF = """
    typedef struct minix_inode {
        uint16 i_mode;
        uint16 i_nlinks;
        uint16 i_uid;
        uint16 i_gid;
        uint32 i_size;
        uint32 i_atime;
        uint32 i_mtime;
        uint32 i_ctime;
        uint32 i_zone[10];
    } minix_inode;
"""

SUPERBLOCK_V1_CDEF = """
    typedef struct minix_super_block {
        uint16 s_ninodes;           /* number of inodes */
        uint16 s_nzones;            /* number of zones (v1 only) */
        uint16 s_imap_blocks;       /* inode map size (in blocks) */
        uint16 s_zmap_blocks;       /* zone map size (in blocks) */
        uint16 s_firstdatazone;     /* first zone containing file data */
        uint16 s_log_zone_size;     /* log_2(zone_size / blocks_size); 0 => zone_size == block_size */
        uint32 s_max_size;          /* max file size */
        uint16 s_magic;             /* minix magic */
        uint16 s_state;             /* mount state */
        uint32 s_zones;             /* number of zones (v2 only) */
    } minix_super_block;
"""

SUPERBLOCK_V3_CDEF = """
    typedef struct minix_super_block {
        uint32 s_ninodes;
        uint16 s_pad0;
        uint16 s_imap_blocks;
        uint16 s_zmap_blocks;
        uint16 s_firstdatazone;
        uint16 s_log_zone_size;
        uint16 s_pad1;
        uint32 s_max_size;
        uint32 s_zones;
        uint16 s_magic;
        uint16 s_pad2;
        uint16 s_blocksize;
        uint8  s_disk_version;
    } minix_super_block;
"""

DIR_V1_CDEF = """
    typedef struct minix_dir_entry {
        uint16 inode;
        char   name[];
    } minix_dir_entry;
"""

DIR_V3_CDEF = """
    typedef struct minix_dir_entry {
        uint32 inode;
        char   name[];
    } minix3_dir_entry;
"""

VERSION_TO_BIG_ENDIAN_MAGIC = {
    1: {0x13_7F, 0x13_8F},
    2: {0x24_68, 0x24_78},
    3: {0x4D_5A},
}
VERSION_TO_MAGIC_OFFSET = {
    1: 0x10,
    2: 0x10,
    3: 0x18,
}

VERSION_TO_C_DEFINITIONS = {
    1: INODE_V1_CDEF + SUPERBLOCK_V1_CDEF + DIR_V1_CDEF,
    2: INODE_V2_CDEF + SUPERBLOCK_V1_CDEF + DIR_V1_CDEF,
    3: INODE_V2_CDEF + SUPERBLOCK_V3_CDEF + DIR_V3_CDEF,
}


class MinixFS:
    def __init__(self, file: File, version: int, c_definitions: str):
        self.file = file
        self.version = version
        self.struct_parser = StructParser(c_definitions)
        self.file.seek(SUPERBLOCK_OFFSET, io.SEEK_SET)
        self.endianness = get_endianness(file, version)
        self.superblock = self.struct_parser.parse(HEADER_STRUCT, file, self.endianness)
        block_size = get_block_size(self.superblock)
        imap_offset = 2 * block_size
        zmap_offset = imap_offset + self.superblock.s_imap_blocks * block_size
        self.inode_offset = zmap_offset + self.superblock.s_zmap_blocks * block_size
        self.zone_size = (block_size << self.superblock.s_log_zone_size) & 0xFF_FF_FF_FF
        self.inode_size = self.struct_parser.cparser_le.minix_inode.size
        self.zone_ptr_size = self.struct_parser.cparser_le.minix_inode.fields[
            "i_zone"
        ].type.type.size
        dirent_inode_size = self.struct_parser.cparser_le.minix_dir_entry.fields[
            "inode"
        ].type.size
        self.dir_entry_size = self._get_name_len() + dirent_inode_size

    def _get_name_len(self) -> int:
        lower_magic: int = self.superblock.s_magic & 0x00FF
        if lower_magic in {0x7F, 0x68}:
            return 14  # v1/v2
        if lower_magic in {0x8F, 0x78}:
            return 30  # v1/v2
        if lower_magic == 0x5A:
            return 60  # v3
        raise InvalidInputFormat(f"Invalid magic: {self.superblock.s_magic:x}")

    def _read_zone_data(self, zone_index: int) -> bytes:
        self.file.seek(zone_index * self.zone_size, io.SEEK_SET)
        return self.file.read(self.zone_size)

    def _get_zone_pointers(self, zone_index: int) -> list[int]:
        data = self._read_zone_data(zone_index)
        ptr_fmt = "H" if self.zone_ptr_size == 2 else "I"
        count = self.zone_size // self.zone_ptr_size
        endianness_fmt = "<" if self.endianness == Endian.LITTLE else ">"
        return list(struct.unpack(f"{endianness_fmt}{count}{ptr_fmt}", data))

    def _stream_file_data(self, inode: MinixInode) -> Iterator[bytes]:
        remaining = inode.i_size
        for zone_data in self._iter_zones(inode.i_zone):
            chunk = zone_data[:remaining]
            remaining -= len(chunk)
            yield chunk
            if remaining <= 0:
                break

    def _iter_zones(self, zones: list[int]) -> Iterator[bytes]:
        # Data zones are a bit complicated. See e.g. https://osblog.stephenmarz.com/ch10.html
        # for a more detailed explanation.
        yield from self._read_zones(zones[:7])  # direct zones 0-6:
        if zones[7] != 0:  # indirect zone 7:
            yield from self._read_zones([zones[7]], 1)
        if zones[8] != 0:  # doubly indirect zone 8:
            yield from self._read_zones([zones[8]], 2)
        if len(zones) == 10 and zones[9] != 0:  # triply indirect zone 9 (V2/V3 only)
            yield from self._read_zones([zones[9]], 3)

    def _read_zones(
        self, zone_index_list: list[int], indirectness: int = 0
    ) -> Iterator[bytes]:
        for index in zone_index_list:
            if index == 0:
                break
            if indirectness > 0:
                zone_pointers = self._get_zone_pointers(index)
                yield from self._read_zones(zone_pointers, indirectness - 1)
            else:
                yield self._read_zone_data(index)

    def _read_directory(self, inode: MinixInode) -> Iterator[MinixDirEntry]:
        for zone_data in self._stream_file_data(inode):
            for i in range(0, len(zone_data), self.dir_entry_size):
                raw_entry = self.struct_parser.parse(
                    "minix_dir_entry",
                    zone_data[i : i + self.dir_entry_size],
                    self.endianness,
                )
                yield MinixDirEntry(inode=raw_entry.inode, name=raw_entry.name)

    def _read_inode(self, index: int) -> MinixInode:
        if not 1 <= index <= self.superblock.s_ninodes:
            raise InvalidInputFormat(f"Invalid inode number: {index}")
        offset = self.inode_offset + (index - 1) * self.inode_size
        self.file.seek(offset, io.SEEK_SET)
        raw_inode = self.struct_parser.parse("minix_inode", self.file, self.endianness)
        if self.version == 1:
            return MinixInode(
                i_mode=raw_inode.i_mode,
                i_nlinks=raw_inode.i_nlinks,
                i_uid=raw_inode.i_uid,
                i_gid=raw_inode.i_gid,
                i_size=raw_inode.i_size,
                i_time=raw_inode.i_time,
                i_atime=None,
                i_mtime=None,
                i_ctime=None,
                i_zone=list(raw_inode.i_zone),
            )
        return MinixInode(
            i_mode=raw_inode.i_mode,
            i_nlinks=raw_inode.i_nlinks,
            i_uid=raw_inode.i_uid,
            i_gid=raw_inode.i_gid,
            i_size=raw_inode.i_size,
            i_time=None,
            i_atime=raw_inode.i_atime,
            i_mtime=raw_inode.i_mtime,
            i_ctime=raw_inode.i_ctime,
            i_zone=list(raw_inode.i_zone),
        )

    def extract(
        self,
        fs: FileSystem,
        inode=None,
        path: Path = Path(),
        inode_index: int = ROOT_INODE_INDEX,
        visited_dirs: set[int] | None = None,
    ):
        if visited_dirs is None:
            visited_dirs = set()

        if not inode:
            try:
                inode = self._read_inode(inode_index)
                if inode is None:
                    raise InvalidInputFormat("Root inode is empty")
                if not stat.S_ISDIR(inode.i_mode):
                    raise InvalidInputFormat("Root entries should be directories")
            except EOFError as error:
                raise InvalidInputFormat("File system is empty") from error

        if inode_index in visited_dirs:
            return
        visited_dirs.add(inode_index)
        self.walk_inode(fs, inode, path, visited_dirs)

    def walk_inode(
        self, fs: FileSystem, inode: MinixInode, path: Path, visited_dirs: set[int]
    ):
        for entry in self._read_directory(inode):
            if entry.name in (b".", b"..") or entry.inode < 1:
                continue
            entry_path = path / entry.name.decode("utf-8", errors="replace")
            entry_inode = self._read_inode(entry.inode)

            if stat.S_ISREG(entry_inode.i_mode):
                fs.write_chunks(entry_path, self._stream_file_data(entry_inode))

            elif stat.S_ISLNK(entry_inode.i_mode):
                contents = b"".join(self._stream_file_data(entry_inode))
                link_target = contents.decode("utf-8", errors="replace")
                fs.create_symlink(Path(link_target), entry_path)

            elif stat.S_ISFIFO(entry_inode.i_mode):
                fs.mkfifo(entry_path, mode=entry_inode.i_mode)

            elif stat.S_ISCHR(entry_inode.i_mode) or stat.S_ISBLK(entry_inode.i_mode):
                fs.mknod(entry_path, mode=entry_inode.i_mode)

            elif stat.S_ISDIR(entry_inode.i_mode):
                fs.mkdir(entry_path, parents=True, exist_ok=True)
                self.extract(fs, entry_inode, entry_path, entry.inode, visited_dirs)


class MinixFSExtractor(Extractor):
    def __init__(self, version: int):
        self.version = version
        self.c_definitions = self._get_c_definitions()

    def _get_c_definitions(self) -> str:
        cdefs = VERSION_TO_C_DEFINITIONS.get(self.version)
        if not cdefs:
            raise ValueError(f"Unsupported MINIX FS version: {self.version}")
        return cdefs

    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            minix = MinixFS(file, self.version, self.c_definitions)
            minix.extract(fs)


def get_endianness(file: File, version: int) -> Endian:
    offset = VERSION_TO_MAGIC_OFFSET[version]
    start = file.tell()
    file.seek(start + offset, io.SEEK_SET)
    magic_bytes = file.read(2)
    file.seek(start, io.SEEK_SET)
    if len(magic_bytes) < 2:
        raise InvalidInputFormat("Not enough bytes to read MINIX magic.")
    magic_be = int.from_bytes(magic_bytes, byteorder="big", signed=False)
    magic_le = int.from_bytes(magic_bytes, byteorder="little", signed=False)
    magics = VERSION_TO_BIG_ENDIAN_MAGIC[version]
    if magic_be in magics:
        return Endian.BIG
    if magic_le in magics:
        return Endian.LITTLE
    raise InvalidInputFormat(f"Invalid MINIX magic: 0x{magic_be:04x}")


def get_block_size(superblock) -> int:
    return getattr(superblock, "s_blocksize", STATIC_BLOCK_SIZE)


class _MinixFSHandlerBase(StructHandler):
    VERSION = 1
    HEADER_STRUCT = "minix_super_block"
    PATTERN_MATCH_OFFSET = -SUPERBLOCK_OFFSET

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        version = getattr(cls, "VERSION", None)
        if version:
            cls.EXTRACTOR = MinixFSExtractor(version=version)
            cls.C_DEFINITIONS = VERSION_TO_C_DEFINITIONS[version]

    def _get_zone_count(self, header) -> int:
        return header.s_nzones

    def validate_header(  # noqa: C901
        self, header, file: File, block_size: int, start_offset: int = 0
    ) -> None:
        if header.s_ninodes < 1:
            raise InvalidInputFormat("Invalid inode count")
        if header.s_imap_blocks < 1:
            raise InvalidInputFormat("Invalid inode map block count")
        if header.s_zmap_blocks < 1:
            raise InvalidInputFormat("Invalid zone map block count")
        if header.s_max_size == 0:
            raise InvalidInputFormat("Invalid max file size")
        # according to https://www.minix3.org/doc/A-312.html valid blocksizes for v3 are 1, 2, 4 and 8 KiB
        if self.VERSION == 3 and header.s_blocksize not in {
            2**x for x in range(10, 14)
        }:
            raise InvalidInputFormat("Invalid block size")
        if header.s_log_zone_size > 10:
            # The default log_zone_size is 0 (meaning zone_size == block_size). Though there does not seem to
            # be a hard cap on this value, values larger than 10 (2**10 = 1024 blocks per zone) are not realistic
            raise InvalidInputFormat("Invalid log zone size")
        zone_count = header.s_zones or header.s_nzones
        if zone_count < 1:
            raise InvalidInputFormat("Invalid zone count")

        blocks_per_zone = 2**header.s_log_zone_size
        total_size = zone_count * blocks_per_zone * block_size
        if total_size > file.size() - start_offset:
            raise InvalidInputFormat("larger than the file size")
        inode_size = 32 if self.VERSION == 1 else 64
        inodes_per_block = block_size // inode_size
        first_data_block = (
            2  # boot block + superblock
            + header.s_imap_blocks
            + header.s_zmap_blocks
            + math.ceil(header.s_ninodes / inodes_per_block)
        )
        first_data_zone = math.ceil(first_data_block / blocks_per_zone)
        if header.s_firstdatazone != first_data_zone:
            raise InvalidInputFormat("Invalid first data zone")

        if self._get_zone_count(header) == 0:
            raise InvalidInputFormat("Invalid zone count")

    def is_valid_header(self, header, file: File, block_size: int) -> bool:
        try:
            self.validate_header(header, file, block_size)
        except InvalidInputFormat:
            return False
        return True

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset + SUPERBLOCK_OFFSET, io.SEEK_SET)
        endianness = get_endianness(file, self.VERSION)
        superblock = self.parse_header(file, endianness)

        # TODO: should probably be moved to MinixFS and rename MinixFS to MinixFSParser,
        # with a get_end_offset() function.

        block_size = get_block_size(superblock)
        self.validate_header(superblock, file, block_size, start_offset)

        zone_size = 2**superblock.s_log_zone_size
        zone_count = self._get_zone_count(superblock)
        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + zone_count * zone_size * block_size,
        )


class MinixFSv1Handler(_MinixFSHandlerBase):
    NAME = "minix_fs_v1"
    PATTERNS = [
        # the magic comes at offset 0x10 in the header
        # the field that comes after this (s_state) indicates the FS state (1 -> valid; 2 -> error).
        # A value of 0 has been seen in the wild, but is not documented.
        # There are two variants with the only difference being the maximum name length (0x7f -> 14; 0x8f -> 30)
        HexString("[16] (7f | 8f) 13 (00 | 01 | 02) 00 [2] 00 00 00 00"),  # LE
        HexString("[16] 13 (7f | 8f) 00 (00 | 01 | 02) [2] 00 00 00 00"),  # BE
    ]
    VERSION = 1

    DOC = HandlerDoc(
        name="MINIX FS (v1)",
        description="MINIX FS is a simple file system format designed as the filesystem of MINIX. MINIX is a UNIX-like operating system, originally developed by Andrew S. Tanenbaum for educational purposes.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Official website",
                url="https://www.minix3.org/",
            ),
            Reference(
                title="Linux headers (minix_fs.h)",
                url="https://github.com/torvalds/linux/blob/master/include/uapi/linux/minix_fs.h",
            ),
            Reference(
                title="Official tool for creating MINIX filesystems",
                url="https://github.com/Stichting-MINIX-Research-Foundation/minix/tree/master/minix/usr.sbin/mkfs.mfs",
            ),
        ],
        limitations=[],
    )


class MinixFSv2Handler(MinixFSv1Handler):
    NAME = "minix_fs_v2"
    PATTERNS = [
        # v2 also has two variants regarding the maximum name length (0x68 -> 14; 0x78 -> 30)
        HexString("[16] (68 | 78) 24 (00 | 01 | 02) 00 [4] 00 00 00 00"),  # LE
        HexString("[16] 24 (68 | 78) 00 (00 | 01 | 02) [4] 00 00 00 00"),  # BE
    ]
    VERSION = 2

    def _get_zone_count(self, header) -> int:
        return header.s_zones

    DOC = HandlerDoc(
        name="MINIX FS (v2)",
        description="MINIX FS is a simple file system format designed as the filesystem of MINIX. MINIX is a UNIX-like operating system, originally developed by Andrew S. Tanenbaum for educational purposes.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Official website",
                url="https://www.minix3.org/",
            ),
            Reference(
                title="Linux headers (minix_fs.h)",
                url="https://github.com/torvalds/linux/blob/master/include/uapi/linux/minix_fs.h",
            ),
            Reference(
                title="Official tool for creating MINIX filesystems",
                url="https://github.com/Stichting-MINIX-Research-Foundation/minix/tree/master/minix/usr.sbin/mkfs.mfs",
            ),
        ],
        limitations=[],
    )


class MinixFSv3Handler(MinixFSv2Handler):
    NAME = "minix_fs_v3"
    PATTERNS = [
        HexString("[4] 00 00 [18] 5a 4d 00 00"),  # LE
        HexString("[4] 00 00 [18] 4d 5a 00 00"),  # BE
    ]
    VERSION = 3

    DOC = HandlerDoc(
        name="MINIX FS (v3)",
        description="MINIX FS is a simple file system format designed as the filesystem of MINIX. MINIX is a UNIX-like operating system, originally developed by Andrew S. Tanenbaum for educational purposes.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Official website",
                url="https://www.minix3.org/",
            ),
            Reference(
                title="Linux headers (minix_fs.h)",
                url="https://github.com/torvalds/linux/blob/master/include/uapi/linux/minix_fs.h",
            ),
            Reference(
                title="Official tool for creating MINIX filesystems",
                url="https://github.com/Stichting-MINIX-Research-Foundation/minix/tree/master/minix/usr.sbin/mkfs.mfs",
            ),
        ],
        limitations=[],
    )



"""Handler for gzip compression format.

It is based on standard documented at
https://datatracker.ietf.org/doc/html/rfc1952.

The handler will create valid chunks for each gzip compressed stream
instead of concatenating sequential streams into an overall
ValidChunk.

We monkey patched Python builtin gzip's _GzipReader read() function to
stop reading as soon as it reach the EOF marker of the current gzip
stream.  This is a requirement for unblob given that streams can be
malformed and followed by garbage/random content that triggers
BadGzipFile errors when gzip library tries to read the next stream
header.
"""

import gzip
import io
import struct
import zlib
from pathlib import Path

from structlog import get_logger

from unblob.extractors import Command
from unblob.extractors.command import MultiFileCommand
from unblob.models import Extractor

from ...file_utils import InvalidInputFormat
from ...models import (
    DirectoryExtractor,
    DirectoryHandler,
    ExtractResult,
    File,
    Glob,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    MultiFile,
    Reference,
    ValidChunk,
)
from ._gzip_reader import SingleMemberGzipReader

logger = get_logger()

GZIP2_CRC_LEN = 4
GZIP2_SIZE_LEN = 4
GZIP2_FOOTER_LEN = GZIP2_CRC_LEN + GZIP2_SIZE_LEN

FLAG_EXTRA = 4
FLAG_NAME = 8


def get_gzip_embedded_name(path: Path) -> str:
    name = b""
    with path.open("rb") as file:
        # skip magic bytes and method
        file.read(2)
        (_method, flag, _last_mtime) = struct.unpack("<BBIxx", file.read(8))

        if flag & FLAG_EXTRA:
            # Read & discard the extra field, if present
            [extra_len] = struct.unpack("<H", file.read(2))
            file.seek(extra_len, io.SEEK_CUR)

        if flag & FLAG_NAME:
            # Read and discard a null-terminated string containing the filename
            while True:
                s = file.read(1)
                if not s or s == b"\000":
                    break
                name += s

    # return a valid, safe name without directories!
    try:
        return Path(name.decode("utf-8")).name
    except UnicodeDecodeError:
        return ""


class GZIPExtractor(Extractor):
    def get_dependencies(self) -> list[str]:
        return ["7z"]

    def extract(self, inpath: Path, outdir: Path) -> ExtractResult | None:
        name = get_gzip_embedded_name(inpath) or "gzip.uncompressed"
        extractor = Command("7z", "x", "-y", "{inpath}", "-so", stdout=name)
        return extractor.extract(inpath, outdir)


class MultiGZIPExtractor(DirectoryExtractor):
    def get_dependencies(self) -> list[str]:
        return ["7z"]

    def extract(self, paths: list[Path], outdir: Path) -> ExtractResult | None:
        name = get_gzip_embedded_name(paths[0]) or "gzip.uncompressed"
        extractor = MultiFileCommand(
            "7z", "x", "-p", "-y", "{inpath}", "-so", stdout=name
        )
        return extractor.extract(paths, outdir)


class GZIPHandler(Handler):
    NAME = "gzip"

    EXTRACTOR = GZIPExtractor()

    PATTERNS = [
        HexString(
            """
            // ID1
            1F
            // ID2
            8B
            // compression method (0x8 = DEFLATE)
            08
            // flags, 00011111 (0x1f) is the highest since the first 3 bits are reserved
            (
                00 | 01 | 02 | 03 | 04 | 05 | 06 | 07 | 08 | 09 |
                0A | 0B | 0C | 0D | 0E | 0F | 10 | 11 | 12 | 13 |
                14 | 15 | 16 | 17 | 18 | 19 | 1A | 1B | 1C | 1D | 1E
            )
            // unix time (uint32) + eXtra FLags (2 or 4 per RFC1952 2.3.1)
            // we accept any value because the RFC is not followed by some samples
            [5]
            // Operating System (0-13, or 255 per RFC1952 2.3.1)
            (
                00 | 01 | 02 | 03 | 04 | 05 | 06 | 07 | 08 | 09 | 0A | 0B | 0C | 0D | FF
            )
        """
        )
    ]

    DOC = HandlerDoc(
        name="GZIP",
        description="GZIP is a compressed file format that uses the DEFLATE algorithm and includes metadata such as original file name and modification time. It is commonly used for efficient file storage and transfer.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="GZIP File Format Specification",
                url="https://datatracker.ietf.org/doc/html/rfc1952",
            ),
            Reference(
                title="GZIP Wikipedia",
                url="https://en.wikipedia.org/wiki/Gzip",
            ),
        ],
        limitations=[],
    )

    def _read_member_end_offset(self, file: File) -> int | None:
        fp = SingleMemberGzipReader(file)
        if not fp.read_header():
            return None

        try:
            fp.read_until_eof()
        except (gzip.BadGzipFile, zlib.error) as e:
            raise InvalidInputFormat from e

        file.seek(GZIP2_FOOTER_LEN - len(fp.unused_data), io.SEEK_CUR)

        return file.tell()

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        end_offset = self._read_member_end_offset(file)
        if end_offset is None:
            return None

        while end_offset < file.size():
            try:
                next_end_offset = self._read_member_end_offset(file)
            except (gzip.BadGzipFile, InvalidInputFormat):
                # Stop at the last valid member if a later member-like sequence is malformed.
                break

            if next_end_offset is None:
                break

            end_offset = next_end_offset

        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )


class MultiVolumeGzipHandler(DirectoryHandler):
    NAME = "multi-gzip"
    EXTRACTOR = MultiGZIPExtractor()

    PATTERN = Glob("*.gz.*")

    DOC = HandlerDoc(
        name="GZIP (multi-volume)",
        description="GZIP is a compressed file format that uses the DEFLATE algorithm and includes metadata such as original file name and modification time. It is commonly used for efficient file storage and transfer.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="GZIP File Format Specification",
                url="https://datatracker.ietf.org/doc/html/rfc1952",
            ),
            Reference(
                title="GZIP Wikipedia",
                url="https://en.wikipedia.org/wiki/Gzip",
            ),
        ],
        limitations=[],
    )

    def is_valid_gzip(self, path: Path) -> bool:
        try:
            file = File.from_path(path)
        except ValueError:
            return False

        with file as f:
            try:
                fp = SingleMemberGzipReader(f)
                if not fp.read_header():
                    return False
            except gzip.BadGzipFile:
                return False
        return True

    def calculate_multifile(self, file: Path) -> MultiFile | None:
        paths = sorted(
            [p for p in file.parent.glob(f"{file.stem}.*") if p.resolve().exists()]
        )

        # we 'discard' paths that are not the first in the ordered list,
        # otherwise we will end up with colliding reports, one for every
        # path in the list.
        if not paths or file != paths[0]:
            return None

        if self.is_valid_gzip(file):
            files_size = sum(path.stat().st_size for path in paths)
            logger.debug(
                "Multi-volume files", paths=paths, files_size=files_size, _verbosity=2
            )

            return MultiFile(
                name=paths[0].stem,
                paths=paths,
            )
        return None



import io
import stat
import struct
from collections.abc import Iterator
from pathlib import Path

from unblob.file_utils import (
    Endian,
    FileSystem,
    InvalidInputFormat,
    StructParser,
)
from unblob.models import (
    Extractor,
    ExtractResult,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

UFS_C_DEFINITION = """
        #define UFS_MAXMNTLEN    512
        #define UFS2_MAXMNTLEN   468
        #define UFS2_MAXVOLLEN   32
        #define UFS_MAXCSBUFS    31
        #define UFS2_NOCSPTRS    28

        struct ufs_csum {
            uint32  cs_ndir;
            uint32  cs_nbfree;
            uint32  cs_nifree;
            uint32  cs_nffree;
        };

        struct ufs2_csum_total {
            uint64  cs_ndir;
            uint64  cs_nbfree;
            uint64  cs_nifree;
            uint64  cs_nffree;
            uint64  cs_numclusters;
            uint64  cs_spare[3];
        };

        struct ufs_timeval {
            uint32  tv_sec;
            uint32  tv_usec;
        };

        struct ufs_super_block {
            union {
                struct {
                    uint32  fs_link;        /* UNUSED */
                } fs_42;
                struct {
                    uint32  fs_state;       /* file system state flag */
                } fs_sun;
            } fs_u0;

            uint32  fs_rlink;               /* UNUSED */
            uint32  fs_sblkno;              /* addr of super-block in filesys */
            uint32  fs_cblkno;              /* offset of cyl-block in filesys */
            uint32  fs_iblkno;              /* offset of inode-blocks in filesys */
            uint32  fs_dblkno;              /* offset of first data after cg */
            uint32  fs_cgoffset;            /* cylinder group offset in cylinder */
            uint32  fs_cgmask;              /* used to calc mod fs_ntrak */
            uint32  fs_time;                /* last time written */
            uint32  fs_size;                /* number of blocks in fs */
            uint32  fs_dsize;               /* number of data blocks in fs */
            uint32  fs_ncg;                 /* number of cylinder groups */
            uint32  fs_bsize;               /* size of basic blocks in fs */
            uint32  fs_fsize;               /* size of frag blocks in fs */
            uint32  fs_frag;                /* number of frags in a block in fs */
            /* configuration parameters */
            uint32  fs_minfree;             /* minimum percentage of free blocks */
            uint32  fs_rotdelay;            /* num of ms for optimal next block */
            uint32  fs_rps;                 /* disk revolutions per second */
            /* fields computed from the others */
            uint32  fs_bmask;               /* blkoff calc of blk offsets */
            uint32  fs_fmask;               /* fragoff calc of frag offsets */
            uint32  fs_bshift;              /* lblkno calc of logical blkno */
            uint32  fs_fshift;              /* numfrags calc number of frags */
            /* configuration parameters */
            uint32  fs_maxcontig;           /* max number of contiguous blks */
            uint32  fs_maxbpg;              /* max number of blks per cyl group */
            /* fields computed from the others */
            uint32  fs_fragshift;           /* block to frag shift */
            uint32  fs_fsbtodb;             /* fsbtodb and dbtofsb shift constant */
            uint32  fs_sbsize;              /* actual size of super block */
            uint32  fs_csmask;              /* csum block offset */
            uint32  fs_csshift;             /* csum block number */
            uint32  fs_nindir;              /* value of NINDIR */
            uint32  fs_inopb;               /* value of INOPB */
            uint32  fs_nspf;                /* value of NSPF */
            /* optimization preference */
            uint32  fs_optim;

            /* fields derived from the hardware */
            union {
                struct {
                    uint32  fs_npsect;      /* # sectors/track including spares */
                } fs_sun;
                struct {
                    uint32  fs_state;       /* file system state time stamp */
                } fs_sunx86;
            } fs_u1;

            uint32  fs_interleave;          /* hardware sector interleave */
            uint32  fs_trackskew;           /* sector 0 skew, per track */
            uint32  fs_id[2];               /* file system id */
            /* sizes determined by number of cylinder groups and their sizes */
            uint32  fs_csaddr;              /* blk addr of cyl grp summary area */
            uint32  fs_cssize;              /* size of cyl grp summary area */
            uint32  fs_cgsize;              /* cylinder group size */
            /* fields derived from the hardware */
            uint32  fs_ntrak;               /* tracks per cylinder */
            uint32  fs_nsect;               /* sectors per track */
            uint32  fs_spc;                 /* sectors per cylinder */
            /* this comes from the disk driver partitioning */
            uint32  fs_ncyl;                /* cylinders in file system */
            /* fields computed from the others */
            uint32  fs_cpg;                 /* cylinders per group */
            uint32  fs_inodes_per_group;    /* inodes per cylinder group */
            uint32  fs_frags_per_group;     /* blocks per group * fs_frag */

            /* this data must be re-computed after crashes */
            struct ufs_csum fs_cstotal;     /* cylinder summary information */

            /* fields cleared at mount time */
            int8    fs_fmod;                /* super block modified flag */
            int8    fs_clean;               /* file system is clean flag */
            int8    fs_ronly;               /* mounted read-only flag */
            int8    fs_flags;

            union {
                struct {
                    int8    fs_fsmnt[UFS_MAXMNTLEN];    /* name mounted on */
                    uint32  fs_cgrotor;                 /* last cg searched */
                    uint32  fs_csp[UFS_MAXCSBUFS];      /* list of fs_cs info buffers */
                    uint32  fs_maxcluster;
                    uint32  fs_cpc;                     /* cyl per cycle in postbl */
                    uint16  fs_opostbl[16][8];          /* old rotation block list head */
                } fs_u1;
                struct {
                    int8    fs_fsmnt[UFS2_MAXMNTLEN];   /* name mounted on */
                    uint8   fs_volname[UFS2_MAXVOLLEN]; /* volume name */
                    uint64  fs_swuid;                   /* system-wide uid */
                    uint32  fs_pad;                     /* due to alignment of fs_swuid */
                    uint32  fs_cgrotor;                 /* last cg searched */
                    uint32  fs_ocsp[UFS2_NOCSPTRS];     /* list of fs_cs info buffers */
                    uint32  fs_contigdirs;              /* # of contiguously allocated dirs */
                    uint32  fs_csp;                     /* cg summary info buffer */
                    uint32  fs_maxcluster;
                    uint32  fs_active;                  /* used by snapshots to track fs */
                    uint32  fs_old_cpc;                 /* cyl per cycle in postbl */
                    uint32  fs_maxbsize;                /* maximum blocking factor permitted */
                    uint64  fs_sparecon64[17];          /* old rotation block list head */
                    uint64  fs_sblockloc;               /* byte offset of standard superblock */
                    struct  ufs2_csum_total fs_cstotal; /* cylinder summary information */
                    struct  ufs_timeval    fs_time;     /* last time written */
                    uint64  fs_size_64;                 /* number of blocks in fs */
                    uint64  fs_dsize;                   /* number of data blocks in fs */
                    uint64  fs_csaddr;                  /* blk addr of cyl grp summary area */
                    uint64  fs_pendingblocks;           /* blocks in process of being freed */
                    uint32  fs_pendinginodes;           /* inodes in process of being freed */
                } fs_u2;
            } fs_u11;

            union {
                struct {
                    uint32  fs_sparecon[53];            /* reserved for future constants */
                    uint32  fs_reclaim;
                    uint32  fs_sparecon2[1];
                    uint32  fs_state;                   /* file system state time stamp */
                    uint32  fs_qbmask[2];               /* ~usb_bmask */
                    uint32  fs_qfmask[2];               /* ~usb_fmask */
                } fs_sun;
                struct {
                    uint32  fs_sparecon[53];            /* reserved for future constants */
                    uint32  fs_reclaim;
                    uint32  fs_sparecon2[1];
                    uint32  fs_npsect;                  /* # sectors/track including spares */
                    uint32  fs_qbmask[2];               /* ~usb_bmask */
                    uint32  fs_qfmask[2];               /* ~usb_fmask */
                } fs_sunx86;
                struct {
                    uint32  fs_sparecon[50];            /* reserved for future constants */
                    uint32  fs_contigsumsize;           /* size of cluster summary array */
                    uint32  fs_maxsymlinklen;           /* max length of an internal symlink */
                    uint32  fs_inodefmt;                /* format of on-disk inodes */
                    uint32  fs_maxfilesize[2];          /* max representable file size */
                    uint32  fs_qbmask[2];               /* ~usb_bmask */
                    uint32  fs_qfmask[2];               /* ~usb_fmask */
                    uint32  fs_state;                   /* file system state time stamp */
                } fs_44;
            } fs_u2_arch;

            uint32  fs_postblformat;        /* format of positional layout tables */
            uint32  fs_nrpos;               /* number of rotational positions */
            uint32  fs_postbloff;           /* rotation block list head */
            uint32  fs_rotbloff;            /* blocks for each rotation */
            uint32  fs_magic;               /* magic number */
            uint8   fs_space[1];            /* list of blocks for each rotation */
} ufs_superblock_t;

        #define UFS_NDADDR  12
        #define UFS_NINDIR  3

        /* UFS1 FreeBSD & Solaris */
        struct ufs1_inode {
            uint16	mode;		/*  0x0 */
            uint16	nlink;		/*  0x2 */
            union {
                struct {
                    uint16	suid;	/*  0x4 */
                    uint16	sgid;	/*  0x6 */
                } oldids;
                uint32	inumber;		/*  0x4 lsf: inode number */
                uint32	author;		/*  0x4 GNU HURD: author */
            } u1;
            uint64	size;		/*  0x8 */
            struct ufs_timeval atime;	/* 0x10 access */
            struct ufs_timeval mtime;	/* 0x18 modification */
            struct ufs_timeval ctime;	/* 0x20 creation */
            union {
                struct {
                    uint32	direct_blocks[UFS_NDADDR];/* 0x28 data blocks */
                    uint32	indirect_blocks[UFS_NINDIR];/* 0x58 indirect blocks */
                } addr;
                uint8	symlink[4*(UFS_NDADDR+UFS_NINDIR)];/* 0x28 fast symlink */
            } u2;
            uint32	flags;		/* 0x64 immutable, append-only... */
            uint32	blocks;		/* 0x68 blocks in use */
            uint32	gen;			/* 0x6c like ext2 i_version, for NFS support */
            union {
                struct {
                    uint32	shadow;	/* 0x70 shadow inode with security data */
                    uint32	uid;		/* 0x74 long EFT version of uid */
                    uint32	gid;		/* 0x78 long EFT version of gid */
                    uint32	oeftflag;	/* 0x7c reserved */
                } sun;
                struct {
                    uint32	uid;		/* 0x70 File owner */
                    uint32	gid;		/* 0x74 File group */
                    uint32	spare[2];	/* 0x78 reserved */
                } bsd44;
                struct {
                    uint32	uid;		/* 0x70 */
                    uint32	gid;		/* 0x74 */
                    uint16	modeh;	/* 0x78 mode high bits */
                    uint16	spare;	/* 0x7A unused */
                    uint32	trans;	/* 0x7c filesystem translator */
                } hurd;
            } u3;
        } ufs1_inode_t;

        #define UFS_NXADDR  2

        /* ---- UFS2 on-disk inode (256 bytes) ---- */
        struct ufs2_inode {
            uint16     mode;        /*   0: IFMT, permissions; see below. */
            uint16     nlink;       /*   2: File link count. */
            uint32     uid;         /*   4: File owner. */
            uint32     gid;         /*   8: File group. */
            uint32     blksize;     /*  12: Inode blocksize. */
            uint64     size;        /*  16: File byte count. */
            uint64     blocks;      /*  24: Bytes actually held. */
            uint64   atime;       /*  32: Last access time. */
            uint64   mtime;       /*  40: Last modified time. */
            uint64   ctime;       /*  48: Last inode change time. */
            uint64   birthtime;   /*  56: Inode creation time. */
            uint32     mtimensec;   /*  64: Last modified time. */
            uint32     atimensec;   /*  68: Last access time. */
            uint32     ctimensec;   /*  72: Last inode change time. */
            uint32     birthnsec;   /*  76: Inode creation time. */
            uint32     gen;         /*  80: Generation number. */
            uint32     kernflags;   /*  84: Kernel flags. */
            uint32     flags;       /*  88: Status flags (chflags). */
            uint32     extsize;     /*  92: External attributes block. */
            uint64     extb[UFS_NXADDR];/*  96: External attributes block. */
            union {
                struct {
                    uint64     direct_blocks[UFS_NDADDR]; /* 112: Direct disk blocks. */
                    uint64     indirect_blocks[UFS_NINDIR];/* 208: Indirect disk blocks.*/
                } addr;
            uint8	symlink[2*4*(UFS_NDADDR+UFS_NINDIR)];/* 0x28 fast symlink */
            } u2;
            uint64     spare[3];    /* 232: Reserved; currently unused */
        } ufs2_inode_t;

        /* New OpenBSD & FreeBSD */

        struct ufs_dirent {
            uint32  d_ino;  /* inode number of this entry */
            uint16  d_reclen; /* length of this entry */
            uint8   d_type; /* file type */
            uint8   d_namlen;
            char    d_name[d_namlen];
        };

        /* Solaris & old xBSD (no type field) */
        struct old_ufs_dirent {
            uint32 d_ino;
            uint16 d_reclen;
            uint16 d_namlen;
            char d_name[d_namlen];
        }

"""

MAGIC_OFFSET = 0x55C  #  relative to SB_OFFSET
UFS_ROOT_INO = 2
DELETED_INO = 0
MAX_BLOCK_SIZE = 65536  # FreeBSD MAXBSIZE (sys/sys/param.h)


class UFSParser:
    INODE_STRUCT: str
    INODE_SIZE: int
    PTR_SIZE: int
    DIRENT_STRUCT: str = "ufs_dirent"

    def __init__(self, file: File, sb_offset: int):
        self.file = file
        self._struct_parser = StructParser(UFS_C_DEFINITION)
        self.file.seek(sb_offset, io.SEEK_SET)
        self.super_block = self._struct_parser.parse(
            "ufs_superblock_t", self.file, Endian.LITTLE
        )

    def walk_extract(self, fs: FileSystem, ino_num: int, path: Path):  # noqa: C901
        inode = self.read_inode(ino_num)
        file_type = stat.S_IFMT(inode.mode)

        match file_type:
            case stat.S_IFDIR:
                fs.mkdir(path, exist_ok=True)
                for child_ino, name in self.parse_dentries(inode):
                    if name in (".", ".."):
                        continue
                    self.walk_extract(fs, child_ino, path / name)
            case stat.S_IFREG:
                fs.write_chunks(path, self.read_file_content(inode))
            case stat.S_IFLNK:
                fs.create_symlink(src=Path(self.read_symlink(inode)), dst=path)
            case stat.S_IFIFO:
                fs.mkfifo(path)
            case stat.S_IFSOCK:
                fs.mknod(path, mode=inode.mode)
            case stat.S_IFCHR | stat.S_IFBLK:
                fs.mknod(path, mode=inode.mode, device=self.get_direct_blocks(inode)[0])

    def parse_dentries(self, inode) -> Iterator[tuple[int, str]]:
        for chunk in self.read_file_content(inode):
            offset = 0
            while offset < len(chunk):
                entry = self._struct_parser.parse(
                    self.DIRENT_STRUCT, chunk[offset:], Endian.LITTLE
                )
                # d_reclen == 0 means end of valid entries in this block
                if entry.d_reclen == 0:
                    break
                offset += entry.d_reclen
                if entry.d_ino == DELETED_INO:
                    continue
                yield (
                    entry.d_ino,
                    entry.d_name.decode("utf-8", errors="replace"),
                )

    def read_file_content(self, inode) -> Iterator[bytes]:
        remaining = inode.size
        for chunk in self.read_direct_blocks(inode):
            to_read = min(len(chunk), remaining)
            yield chunk[:to_read]
            remaining -= to_read
            if remaining <= 0:
                return
        for chunk in self.read_indirect_blocks(inode, remaining):
            yield chunk

    def read_direct_blocks(self, inode) -> Iterator[bytes]:
        for fragment_index in self.get_direct_blocks(inode):
            if fragment_index == 0:
                # Sparse file: unallocated block reads as zeroes
                yield b"\x00" * self.super_block.fs_bsize
            else:
                self.file.seek(self.frag_to_offset(fragment_index), io.SEEK_SET)
                yield self.file.read(self.super_block.fs_bsize)

    def read_indirect_blocks(self, inode, remaining: int) -> Iterator[bytes]:  # noqa: C901
        indirect_blocks = self.get_indirect_blocks(inode)
        for level, fragment_index in enumerate(indirect_blocks, start=1):
            if fragment_index == 0 or remaining <= 0:
                break
            # levels: single=1, double=2, triple=3
            indexes = [fragment_index]
            for _ in range(level):
                next_indexes = []
                for idx in indexes:
                    if idx != 0:
                        next_indexes.extend(self.read_block_pointers(idx))
                indexes = next_indexes
            for data_index in indexes:
                if remaining <= 0:
                    return
                to_read = min(self.super_block.fs_bsize, remaining)
                if data_index == 0:
                    # Sparse file: unallocated block reads as zeroes
                    yield b"\x00" * to_read
                else:
                    self.file.seek(self.frag_to_offset(data_index), io.SEEK_SET)
                    yield self.file.read(to_read)
                remaining -= to_read

    def read_block_pointers(self, fragment_index: int) -> list[int]:
        self.file.seek(self.frag_to_offset(fragment_index), io.SEEK_SET)
        data = self.file.read(self.super_block.fs_bsize)
        count = self.super_block.fs_bsize // self.PTR_SIZE
        fmt = f"<{count}I" if self.PTR_SIZE == 4 else f"<{count}Q"
        return list(struct.unpack(fmt, data))

    def read_symlink(self, inode) -> str:
        if inode.blocks == 0:
            return bytes(inode.u2.symlink[: inode.size]).decode(
                "utf-8", errors="replace"
            )
        chunk = next(self.read_file_content(inode))
        return chunk[: inode.size].decode("utf-8", errors="replace")

    def read_inode(self, ino_number: int):
        cylinder_group = ino_number // self.super_block.fs_inodes_per_group
        index = ino_number % self.super_block.fs_inodes_per_group
        offset = (
            self.frag_to_offset(
                self.cylinder_group_start(cylinder_group) + self.super_block.fs_iblkno
            )
            + index * self.INODE_SIZE
        )
        self.file.seek(offset, io.SEEK_SET)
        return self._struct_parser.parse(self.INODE_STRUCT, self.file, Endian.LITTLE)

    def frag_to_offset(self, fragment_index: int) -> int:
        """Convert a fragment index to a byte offset."""
        return fragment_index * self.super_block.fs_fsize

    def cylinder_group_start(self, cylinder_group: int) -> int:
        return cylinder_group * self.super_block.fs_frags_per_group

    def get_direct_blocks(self, inode) -> list[int]:
        return inode.u2.addr.direct_blocks

    def get_indirect_blocks(self, inode) -> list[int]:
        return inode.u2.addr.indirect_blocks


class UFS1Parser(UFSParser):
    INODE_STRUCT = "ufs1_inode_t"
    INODE_SIZE = 128
    PTR_SIZE = 4

    def cylinder_group_start(self, cylinder_group: int) -> int:
        # Old UFS1 rotates cylinder group layout to minimize seek time on spinning disks
        return (
            cylinder_group * self.super_block.fs_frags_per_group
            + self.super_block.fs_cgoffset
            * (cylinder_group & ~self.super_block.fs_cgmask)
        )


class UFS2Parser(UFSParser):
    INODE_STRUCT = "ufs2_inode_t"
    INODE_SIZE = 256
    PTR_SIZE = 8


class SolarisUFS1Parser(UFS1Parser):
    DIRENT_STRUCT = "old_ufs_dirent"


class UFSExtractor(Extractor):
    def __init__(self, parser: type[UFSParser], sb_offset: int):
        self.parser = parser
        self.sb_offset = sb_offset

    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as f:
            parser = self.parser(f, self.sb_offset)
            parser.walk_extract(fs, UFS_ROOT_INO, Path("/"))
        return ExtractResult(reports=fs.problems)


class _UFSBaseHandler(StructHandler):
    HEADER_STRUCT = "ufs_superblock_t"
    C_DEFINITIONS = UFS_C_DEFINITION
    EXTRACTOR = None
    SB_OFFSET = 0

    def get_block_size(self, header) -> int:
        raise NotImplementedError("Subclasses must implement this function.")

    def is_valid_header(self, header) -> bool:
        return (
            header.fs_fsize > 0
            and header.fs_bsize > 0
            and header.fs_bsize <= MAX_BLOCK_SIZE
            and header.fs_frag == (header.fs_bsize // header.fs_fsize)
            and self.get_block_size(header) > 0
            and header.fs_ncg > 0
        )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(
            start_offset + self.SB_OFFSET, io.SEEK_SET
        )  # Skip the boot sector to reach the start of UFS superblock
        header = self.parse_header(file)
        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid UFS Header")

        end_offset = start_offset + (self.get_block_size(header) * header.fs_fsize)
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)


class UFS1Handler(_UFSBaseHandler):
    NAME = "ufs1"
    PATTERNS = [
        HexString("( 01 | 02 ) 00 00 00 [8] 54 19 01 00")
    ]  # fs_nrpos +  UFS1 fs_magic + null fs_space
    SB_OFFSET = 0x2000
    EXTRACTOR = UFSExtractor(UFS1Parser, SB_OFFSET)
    PATTERN_MATCH_OFFSET = -SB_OFFSET - (MAGIC_OFFSET - 12)
    DOC = HandlerDoc(
        name="ufs1",
        description="UFS1 (Unix File System 1) is the original UFS implementation supported by Unix-like operating systems such as FreeBSD and Solaris. It utilizes a hierarchical tree structure and inodes to manage file metadata and data block addresses, with 32-bit block addressing limiting partition sizes to 1TB.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Unix File System Wikipedia",
                url="https://en.wikipedia.org/wiki/Unix_File_System",
            )
        ],
        limitations=[],
    )

    def get_block_size(self, header) -> int:
        return header.fs_size


class UFS2Handler(_UFSBaseHandler):
    NAME = "ufs2"
    PATTERNS = [HexString("19 01 54 19 00")]  # UFS2 fs_magic + null fs_space
    SB_OFFSET = 0x10000
    PATTERN_MATCH_OFFSET = -SB_OFFSET - MAGIC_OFFSET
    EXTRACTOR = UFSExtractor(UFS2Parser, SB_OFFSET)
    DOC = HandlerDoc(
        name="ufs2",
        description="UFS2 (Unix File System 2) is an extended version of UFS1 supported by Unix-like operating systems such as FreeBSD. It introduces 64-bit block addressing, extended file attributes, and improved performance over UFS1, while retaining the hierarchical tree structure and inodes for file metadata and data block management.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Unix File System Wikipedia",
                url="https://en.wikipedia.org/wiki/Unix_File_System",
            )
        ],
        limitations=[],
    )

    def get_block_size(self, header) -> int:
        return header.fs_u11.fs_u2.fs_size_64


class SolarisHandler(_UFSBaseHandler):
    NAME = "solaris_ufs1"
    PATTERNS = [HexString("08 00 00 00 [8] 54 19 01 00")]
    SB_OFFSET = 0x2000
    EXTRACTOR = UFSExtractor(SolarisUFS1Parser, SB_OFFSET)
    PATTERN_MATCH_OFFSET = -SB_OFFSET - (MAGIC_OFFSET - 12)
    DOC = HandlerDoc(
        name="solaris_ufs1",
        description="Solaris UFS1 is the variant of UFS1 used by Oracle Solaris and illumos-based systems such as OpenIndiana and OmniOS. It shares the same overall design as FreeBSD UFS1 but keeps an older on-disk convention inherited from early Unix, making it incompatible with FreeBSD's variant despite the shared magic number.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="Unix File System Wikipedia",
                url="https://en.wikipedia.org/wiki/Unix_File_System",
            ),
            Reference(
                title="Oracle Solaris",
                url="https://en.wikipedia.org/wiki/Oracle_Solaris",
            ),
        ],
        limitations=[],
    )

    def get_block_size(self, header) -> int:
        return header.fs_size



from __future__ import annotations

import gzip
import io
from enum import IntEnum
from pathlib import Path

from unblob.file_utils import Endian, File, FileSystem, InvalidInputFormat, StructParser
from unblob.models import (
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    StructHandler,
    ValidChunk,
)

C_DEFINITIONS = r"""
    typedef struct frm_container_header {
        char    magic[4];       /* *FRM */
        uint32  unknown1;       /* version maybe, only 00 00 00 01 observed */
        uint32  total_length;   /* the size of the whole container (should equal the sum of the section sizes + header_length) */
        uint16  header_length;  /* only 60 00 observed -> size 0x60 = 96 */
        uint16  section_count;  /* only 02 00 observed -> 2 section entries (webserver FS and FW binary) */
        uint8   unknown2[48];   /* optional unknown entries; empty in some samples */
    } frm_container_header_t;

    typedef struct frm_section_entry {
        uint32  type;   /* probably type; values 1 (FW binary) and 2 (webserver FS) observed */
        uint32  offset;
        uint32  length;
        uint32  unknown;
    } frm_section_entry_t;

    typedef struct frm_fs_header {
        char    device_name[32];
        uint8   unknown1[4];        /* version maybe? */
        uint32  timestamp;          /* creation time (UNIX timestamp) */
        uint32  unknown2;
        uint32  unknown3;           /* maybe checksum? */
        uint32  file_table_offset;  /* offset of the file table; only 0x100 observed -> usually starts at 0x160 */
        uint32  file_table_length;  /* size of the file table (= file_header_length * file_count) */
        uint16  file_header_length; /* size of each entry; only 0x40 (64) observed */
        uint16  file_count;         /* the number of table entries (files) */
        uint32  data_length;        /* the length of the file content data */
        uint16  unknown4;
        uint16  file_count2;        /* for some devices (e.g. Nport 5600) this field contains the file count and file_count is 0 */
        uint32  file_table_length2;
        uint32  unknown5;
    } frm_fs_header_t;

    typedef struct frm_file_header {
        char    name[48];
        uint8   unknown[8];    /* could be the creation time */
        uint32  file_length;
        uint32  data_offset;
    } frm_file_header_t;
"""

MAGIC = b"*FRM"
GZIP_MAGIC = bytes.fromhex("1f 8b")


class SectionTypes(IntEnum):
    FW_BINARY = 1
    FILESYSTEM = 2


class MoxaFRMExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            header = self._struct_parser.parse(
                "frm_container_header_t", file, Endian.LITTLE
            )
            sections = self._parse_section_table(file, header.section_count)
            for section in sections:
                self._extract_section(section, file, fs)

    def _parse_section_table(self, file: File, section_count: int):
        return [
            self._struct_parser.parse("frm_section_entry_t", file, Endian.LITTLE)
            for _ in range(section_count)
        ]

    def _extract_section(self, section, file: File, fs: FileSystem):
        match section.type:
            case SectionTypes.FW_BINARY:
                fs.carve(Path("firmware.bin"), file, section.offset, section.length)
            case SectionTypes.FILESYSTEM:
                self._extract_fs_section(file, fs, section.offset)
            case _:
                raise InvalidInputFormat(f"Unknown section type: {section.type}")

    def _extract_fs_section(self, file: File, fs: FileSystem, section_offset: int):
        file.seek(section_offset, io.SEEK_SET)
        fs_header = self._struct_parser.parse("frm_fs_header_t", file, Endian.LITTLE)

        file_table_offset = section_offset + fs_header.file_table_offset
        for index in range(fs_header.file_count or fs_header.file_count2):
            entry_offset = file_table_offset + index * fs_header.file_header_length
            file.seek(entry_offset, io.SEEK_SET)
            entry = self._struct_parser.parse("frm_file_header_t", file, Endian.LITTLE)

            name = bytes(entry.name).rstrip(b"\x00")
            if not name:
                continue

            file.seek(section_offset + entry.data_offset, io.SEEK_SET)
            raw = file.read(entry.file_length)
            self._write_file(fs, Path(name.decode("ascii", errors="replace")), raw)

    @staticmethod
    def _write_file(fs: FileSystem, file_path: Path, file_contents: bytes):
        # some (but usually not all) files are GZIP compressed
        if file_contents[:2] == GZIP_MAGIC:
            file_contents = gzip.decompress(file_contents)
        fs.write_bytes(file_path, file_contents)


class MoxaFRMHandler(StructHandler):
    NAME = "moxa_frm"

    PATTERNS = [HexString("2A 46 52 4D 00 00 00 01 [4] 60 00 02 00")]

    DOC = HandlerDoc(
        name="Moxa FRM",
        description=(
            "Firmware container format used in Moxa firmware (e.g. NPort, MGate and MiiNePort devices)."
        ),
        handler_type=HandlerType.ARCHIVE,
        vendor="Moxa",
        references=[],
        limitations=[],
    )

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "frm_container_header_t"
    EXTRACTOR = MoxaFRMExtractor()

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        header = self.parse_header(file)
        return ValidChunk(
            start_offset=start_offset, end_offset=start_offset + header.total_length
        )



import enum
import io
import zlib
from pathlib import Path

import pyzstd
from lzallright import LZOCompressor

from unblob.file_utils import (
    Endian,
    File,
    FileSystem,
    InvalidInputFormat,
    StructParser,
    iterate_file,
)
from unblob.models import (
    Extractor,
    ExtractResult,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)
from unblob.report import ExtractionProblem

C_DEFINITIONS = """
typedef struct stream_header {
    char     magic[13];   // "btrfs-stream\0"
    uint32   version;
} stream_header_t;

typedef struct cmd_header {
    uint32 data_len;
    uint16 cmd_type;
    uint32 crc32; // use Castagnoli polynomial and the seed 0x0
} cmd_header_t;

typedef struct tlv_header_no_value {
    uint16 tlv_type;
    uint16 tlv_len;
} tlv_header_no_value_t;

typedef struct tlv_header_type_only {
    uint16 tlv_type;
} tlv_header_type_only_t;

typedef struct tlv_header {
    uint16 tlv_type;
    uint16 tlv_len;
    uint64 value;
} tlv_header_t;

typedef struct tlv_header_u32 {
    uint16 tlv_type;
    uint16 tlv_len;
    uint32 value;
} tlv_header_u32_t;

typedef struct tlv_header_path {
    uint16 tlv_type;
    uint16 tlv_len;
    char value[tlv_len];
} tlv_header_path_t;

typedef struct timespec {
    uint64 sec;
    uint32 nsec;
} timespec_t;

typedef struct tlv_header_timespec {
    uint16 tlv_type;
    uint16 tlv_len;
    timespec_t value;
} tlv_header_timespec_t;

typedef struct tlv_header_uuid {
    uint16 tlv_type;
    uint16 tlv_len;
    char value[16];
} tlv_header_uuid_t;

typedef struct mk_cmd {
    tlv_header_path_t path;
    tlv_header_t ino;
} mk_cmd_t;

typedef struct symlink_cmd {
    tlv_header_path_t path;
    tlv_header_t ino;
    tlv_header_path_t path_link;
} symlink_cmd_t;

typedef struct rename_cmd {
    tlv_header_path_t path;
    tlv_header_path_t path_to;
} rename_cmd_t;

typedef struct link_cmd {
    tlv_header_path_t path;
    tlv_header_path_t link_to;
} link_cmd_t;

typedef struct truncate_cmd {
    tlv_header_path_t path;
    tlv_header_t size;
} truncate_cmd_t;

typedef struct mk_special_cmd { // MKNOD, MKSOCK, MKFIFO, in the documentation they don't have this structure
    tlv_header_path_t path;
    tlv_header_t      ino;
    tlv_header_t      rdev;
    tlv_header_t      mode;
} mk_special_cmd_t;

typedef struct rmdir_cmd {
    tlv_header_path_t path;
} rmdir_cmd_t;

typedef struct unlink_cmd {
    tlv_header_path_t path;
} unlink_cmd_t;

typedef struct write_cmd {
    tlv_header_path_t path;
    tlv_header_t offset;
} write_cmd_t;

typedef struct clone_cmd {
    tlv_header_path_t path;
    tlv_header_t offset;
    tlv_header_t clone_len;
    tlv_header_t clone_uuid;
    tlv_header_t clone_ctransid;
    tlv_header_path_t clone_path;
    tlv_header_t clone_offset;
} clone_cmd_t;

typedef struct subvol_cmd {
    tlv_header_path_t   path;
    tlv_header_uuid_t   uuid;
    tlv_header_t        ctransid;
} subvol_cmd_t;

typedef struct snapshot_cmd {
    tlv_header_path_t   path;
    tlv_header_uuid_t   uuid;
    tlv_header_t        ctransid;
    tlv_header_uuid_t   clone_uuid;
    tlv_header_t        clone_ctransid;
} snapshot_cmd_t;

typedef struct chmod_cmd {
    tlv_header_path_t   path;
    tlv_header_t        mode;
} chmod_cmd_t;

typedef struct chown_cmd {
    tlv_header_path_t   path;
    tlv_header_t        uid;
    tlv_header_t        gid;
} chown_cmd_t;

typedef struct utimes_cmd {
    tlv_header_path_t       path;
    tlv_header_timespec_t   atime;
    tlv_header_timespec_t   mtime;
    tlv_header_timespec_t   ctime;
} utimes_cmd_t;

typedef struct utimesV2_cmd {
    tlv_header_path_t       path;
    tlv_header_timespec_t   atime;
    tlv_header_timespec_t   mtime;
    tlv_header_timespec_t   ctime;
    tlv_header_timespec_t   otime;
} utimesV2_cmd_t;

typedef struct set_xattr_cmd {
    tlv_header_path_t   path;
    tlv_header_path_t   xattr_name;
    tlv_header_path_t   xattr_data;
} set_xattr_cmd_t;

typedef struct remove_xattr_cmd {
    tlv_header_path_t   path;
    tlv_header_path_t   xattr_name;
} remove_xattr_cmd_t;

typedef struct update_extent_cmd {
    tlv_header_path_t   path;
    tlv_header_t        file_offset;
    tlv_header_t        size;
} update_extent_cmd_t;


typedef struct fallocate_cmd {
    tlv_header_path_t   path;
    tlv_header_t        fallocate_mode;
    tlv_header_t        file_offset;
    tlv_header_t        size;
} fallocate_cmd_t;

typedef struct fileattr_cmd {
    tlv_header_path_t   path;
    tlv_header_t        fileattr;
} fileattr_cmd_t;

typedef struct btrfs_lzo_header {
    uint32 total_len; // total_len (includes itself)
} btrfs_lzo_header_t;

typedef struct btrfs_lzo_segment {
    uint32 seg_len;
} btrfs_lzo_segment_t;

typedef struct encoded_write_cmd {
    tlv_header_path_t   path;
    tlv_header_t        file_offset;
    tlv_header_t        unencoded_file_len;
    tlv_header_t        unencoded_len;
    tlv_header_t        unencoded_offset;
    tlv_header_u32_t    compression;
    tlv_header_u32_t    encryption; // not implemented yet by btrfs send stream
    // DATA has no tlv_len in v2 — length is implicit (cmd.data_len - bytes consumed so far)
} encoded_write_cmd_t;

"""

STREAM_HEADER_LEN = 17
BTRFS_LZO_PAGE_SIZE = 4096


def makecrc32ctable():
    table = []
    for i in range(256):
        crc = i
        for _ in range(8):
            if crc & 1:
                crc = (
                    crc >> 1
                ) ^ 0x82F63B78  # Castagnoli reversed polynomial (little-endian)
            else:
                crc >>= 1
        table.append(crc)
    return table


CRC32CTABLE = makecrc32ctable()


def calculate_crc32(data: bytes, crc: int = 0) -> int:
    for byte in data:
        crc = (crc >> 8) ^ CRC32CTABLE[(crc ^ byte) & 0xFF]
    return crc


def valid_crc32(file: File, cmd_header) -> bool:
    cmd_header_crc32 = cmd_header.crc32
    cmd_header.crc32 = 0x0

    data_start = file.tell()
    tmp_crc32 = calculate_crc32(cmd_header.dumps())
    for chunk in iterate_file(file, data_start, cmd_header.data_len):
        tmp_crc32 = calculate_crc32(chunk, tmp_crc32)
    if cmd_header_crc32 == tmp_crc32:
        file.seek(data_start, io.SEEK_SET)
        return True
    return False


class CmdType(enum.IntEnum):
    UNSPEC = 0
    SUBVOL = 1
    SNAPSHOT = 2
    MKFILE = 3
    MKDIR = 4
    MKNOD = 5
    MKFIFO = 6
    MKSOCK = 7
    SYMLINK = 8
    RENAME = 9
    LINK = 10
    UNLINK = 11
    RMDIR = 12
    SET_XATTR = 13
    REMOVE_XATTR = 14
    WRITE = 15
    CLONE = 16
    TRUNCATE = 17
    CHMOD = 18
    CHOWN = 19
    UTIMES = 20
    END = 21
    UPDATE_EXTENT = 22
    # V2 commands
    FALLOCATE = 23
    FILEATTR = 24
    ENCODED_WRITE = 25


class CompressionType(enum.IntEnum):
    NONE = 0
    ZLIB = 1
    ZSTD = 2
    LZO = 3


class BTRFSParser:
    def __init__(self, file: File, start_offset: int):
        self._struct_parser = StructParser(C_DEFINITIONS)
        self.file = file
        self.start_offset = start_offset
        self.file.seek(self.start_offset, io.SEEK_SET)
        self.stream_header = self._struct_parser.parse(
            "stream_header_t", self.file, Endian.LITTLE
        )

    def decompress_btrfs_lzo(self, data: bytes) -> bytes:
        # btrfs splits each lzo compressed extent into independent 4 KiB pages so the
        # kernel can read any page without decompressing the whole extent.
        stream = io.BytesIO(data)
        total_len = self._struct_parser.parse(
            "btrfs_lzo_header_t",
            stream,  # pyright: ignore[reportArgumentType]
            Endian.LITTLE,
        ).total_len
        decompressed = []
        while stream.tell() < total_len:
            page_remaining = BTRFS_LZO_PAGE_SIZE - (stream.tell() % BTRFS_LZO_PAGE_SIZE)
            # if <4 bytes remain in the page, the rest is padding.
            if page_remaining < 4:
                stream.seek(page_remaining, io.SEEK_CUR)
                continue
            seg_len = self._struct_parser.parse(
                "btrfs_lzo_segment_t",
                stream,  # pyright: ignore[reportArgumentType]
                Endian.LITTLE,
            ).seg_len
            decompressed.append(
                LZOCompressor.decompress(
                    stream.read(seg_len), output_size_hint=BTRFS_LZO_PAGE_SIZE
                )
            )
        return b"".join(decompressed)

    def replay(self, fs: FileSystem) -> None:
        self.file.seek(self.start_offset + len(self.stream_header), io.SEEK_SET)
        cmd_header = self._struct_parser.parse("cmd_header_t", self.file, Endian.LITTLE)

        if CmdType(cmd_header.cmd_type) == CmdType.SNAPSHOT:
            raise InvalidInputFormat("Incremental BTRFS streams are not supported")

        while cmd_header.cmd_type != CmdType.END:
            if not valid_crc32(self.file, cmd_header):
                fs.record_problem(
                    ExtractionProblem(
                        problem=f"Command type : {cmd_header.cmd_type} has an invalid checksum",
                        resolution="Skipped",
                    )
                )
            else:
                self.replay_command(cmd_header, fs)
            cmd_header = self._struct_parser.parse(
                "cmd_header_t", self.file, Endian.LITTLE
            )

    def replay_command(self, cmd_header, fs: FileSystem) -> None:  # noqa: C901
        match CmdType(cmd_header.cmd_type):
            case CmdType.MKFILE:
                command = self._struct_parser.parse(
                    "mk_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                fs.write_bytes(path, b"")
            case CmdType.MKDIR:
                command = self._struct_parser.parse(
                    "mk_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                fs.mkdir(path, parents=True, exist_ok=True)
            case CmdType.MKNOD | CmdType.MKSOCK:
                command = self._struct_parser.parse(
                    "mk_special_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                fs.mknod(path, mode=command.mode.value, device=command.rdev.value)
            case CmdType.MKFIFO:
                command = self._struct_parser.parse(
                    "mk_special_cmd_t", self.file, Endian.LITTLE
                )
                fs.mkfifo(Path(command.path.value.decode()))
            case CmdType.SYMLINK:
                command = self._struct_parser.parse(
                    "symlink_cmd_t", self.file, Endian.LITTLE
                )
                fs.create_symlink(
                    src=Path(command.path_link.value.decode()),
                    dst=Path(command.path.value.decode()),
                )
            case CmdType.RENAME:
                command = self._struct_parser.parse(
                    "rename_cmd_t", self.file, Endian.LITTLE
                )
                fs.rename(
                    src=Path(command.path.value.decode()),
                    dst=Path(command.path_to.value.decode()),
                )
            case CmdType.LINK:
                command = self._struct_parser.parse(
                    "link_cmd_t", self.file, Endian.LITTLE
                )
                fs.create_hardlink(
                    src=Path(command.link_to.value.decode()),
                    dst=Path(command.path.value.decode()),
                )
            case CmdType.RMDIR:
                command = self._struct_parser.parse(
                    "rmdir_cmd_t", self.file, Endian.LITTLE
                )
                fs.rmdir(Path(command.path.value.decode()))
            case CmdType.UNLINK:
                command = self._struct_parser.parse(
                    "unlink_cmd_t", self.file, Endian.LITTLE
                )
                fs.unlink(Path(command.path.value.decode()))
            case CmdType.SET_XATTR:
                command = self._struct_parser.parse(
                    "set_xattr_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                name = command.xattr_name.value.decode()
                data = command.xattr_data.value
                fs.set_xattr(path, name, data)
            case CmdType.REMOVE_XATTR:
                command = self._struct_parser.parse(
                    "remove_xattr_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                name = command.xattr_name.value.decode()
                fs.remove_xattr(path, name)
            case CmdType.WRITE:
                self.replay_write(fs, cmd_header)
            case CmdType.CLONE:
                self.replay_clone(fs)
            case CmdType.TRUNCATE:
                command = self._struct_parser.parse(
                    "truncate_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                size = command.size.value
                fs.truncate(path, size)
            case CmdType.ENCODED_WRITE:
                self.replay_encoded_write(fs, cmd_header)
            case CmdType.UTIMES:
                if self.stream_header.version == 2:
                    command = self._struct_parser.parse(
                        "utimesV2_cmd_t", self.file, Endian.LITTLE
                    )
                else:
                    command = self._struct_parser.parse(
                        "utimes_cmd_t", self.file, Endian.LITTLE
                    )
                path = Path(command.path.value.decode())
                times = (command.atime.value.sec, command.mtime.value.sec)
                fs.utime(path, times)
            case CmdType.CHMOD:
                command = self._struct_parser.parse(
                    "chmod_cmd_t", self.file, Endian.LITTLE
                )
                path = Path(command.path.value.decode())
                mode = command.mode.value
                fs.chmod(path, mode)
            case (
                CmdType.CHOWN
                | CmdType.UPDATE_EXTENT
                | CmdType.FALLOCATE
                | CmdType.FILEATTR
            ):
                fs.record_problem(
                    ExtractionProblem(
                        problem=f"Unsupported BTRFS stream command: {CmdType(cmd_header.cmd_type).name}",
                        resolution="Skipped",
                    )
                )
                self.file.seek(cmd_header.data_len, io.SEEK_CUR)
            case CmdType.UNSPEC | CmdType.END | CmdType.SUBVOL | CmdType.SNAPSHOT:
                self.file.seek(cmd_header.data_len, io.SEEK_CUR)

    def replay_write(self, fs: FileSystem, cmd_header):
        command = self._struct_parser.parse("write_cmd_t", self.file, Endian.LITTLE)
        path = Path(command.path.value.decode())
        if self.stream_header.version == 2:
            # v2: DATA TLV has no length field, skip the 2-byte type field
            tlv_type = self._struct_parser.parse(
                "tlv_header_type_only_t", self.file, Endian.LITTLE
            )
            data_len = cmd_header.data_len - len(command) - len(tlv_type)
        else:
            # ok to handle this TLV this way, otherwise cstruct reads the whole data in ram
            tlv = self._struct_parser.parse(
                "tlv_header_no_value_t", self.file, Endian.LITTLE
            )
            data_len = tlv.tlv_len

        with fs.open(path, "rb+") as f:
            f.seek(command.offset.value, io.SEEK_SET)
            for chunk in iterate_file(self.file, self.file.tell(), data_len):
                f.write(chunk)

    def replay_encoded_write(self, fs: FileSystem, cmd_header) -> None:
        command = self._struct_parser.parse(
            "encoded_write_cmd_t", self.file, Endian.LITTLE
        )
        path = Path(command.path.value.decode())
        # V2 only: DATA TLV has no tlv_len, skip the 2-byte type field
        tlv_type = self._struct_parser.parse(
            "tlv_header_type_only_t", self.file, Endian.LITTLE
        )
        data_len = cmd_header.data_len - len(command) - len(tlv_type)

        # data_len is bounded by BTRFS_MAX_COMPRESSED (128 KiB); large files are
        # split into multiple encoded_write commands, so reading at once is safe
        data = self.file.read(data_len)
        match CompressionType(command.compression.value):
            case CompressionType.NONE:
                decoded = data
            case CompressionType.LZO:
                decoded = self.decompress_btrfs_lzo(data)
            case CompressionType.ZLIB:
                decoded = zlib.decompressobj().decompress(data)
            case CompressionType.ZSTD:
                decoded = pyzstd.ZstdDecompressor().decompress(data)
            case _:
                raise InvalidInputFormat
        with fs.open(path, "rb+") as f:
            f.seek(command.file_offset.value, io.SEEK_SET)
            f.write(decoded)

    def replay_clone(self, fs: FileSystem):
        command = self._struct_parser.parse("clone_cmd_t", self.file, Endian.LITTLE)
        dst_path = Path(command.path.value.decode())
        dst_offset = command.offset.value
        src_len = command.clone_len.value
        src_path = Path(command.clone_path.value.decode())
        src_offset = command.clone_offset.value
        with fs.open(src_path, "rb+") as src, fs.open(dst_path, "rb+") as dst:
            dst.seek(dst_offset, io.SEEK_SET)
            for chunk in iterate_file(src, src_offset, src_len):  # pyright: ignore[reportArgumentType]
                dst.write(chunk)


class BTRFSStreamExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path) -> ExtractResult:
        fs = FileSystem(outdir)
        with File.from_path(inpath) as file:
            btrfs_parser = BTRFSParser(file, 0)
            btrfs_parser.replay(fs)
        return ExtractResult(reports=fs.problems)


class BTRFSStreamHandler(StructHandler):
    NAME = "btrfs_stream"
    PATTERNS = [HexString("62 74 72 66 73 2d 73 74 72 65 61 6d 00 (01 | 02) 00 00 00")]
    C_DEFINITIONS = C_DEFINITIONS
    EXTRACTOR = BTRFSStreamExtractor()
    HEADER_STRUCT = "stream_header_t"
    DOC = HandlerDoc(
        name="BTRFS Stream",
        description="A BTRFS send stream is a binary format used to transfer btrfs subvolume snapshots between filesystems. It encodes filesystem operations (file creation, directory structure, data writes, metadata) as a sequence of commands that can be replayed by btrfs receive to reconstruct the original snapshot. It supports both full sends (complete snapshot) and incremental sends (diff between two snapshots).",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="BTRFS Stream Official Documentation",
                url="https://btrfs.readthedocs.io/en/latest/dev/dev-send-stream.html",
            ),
        ],
        limitations=["Does not support incremental streams."],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        file.seek(start_offset + STREAM_HEADER_LEN, io.SEEK_SET)
        cmd_header = self.cparser_le.cmd_header_t(file)
        while cmd_header.cmd_type != CmdType.END:
            try:
                CmdType(cmd_header.cmd_type)
            except ValueError as err:
                raise InvalidInputFormat(
                    f"Invalid BTRFS stream command type: {cmd_header.cmd_type}"
                ) from err
            file.seek(cmd_header.data_len, io.SEEK_CUR)
            cmd_header = self.cparser_le.cmd_header_t(file)
        return ValidChunk(start_offset=start_offset, end_offset=file.tell())



"""RAR handler.

RAR Version 4.x
https://codedread.github.io/bitjs/docs/unrar.html.

RAR Version 5.x
https://www.rarlab.com/technote.htm#rarsign
"""

import rarfile
from structlog import get_logger

from unblob.extractors import Command

from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)

logger = get_logger()


class RarHandler(Handler):
    NAME = "rar"

    PATTERNS = [
        HexString(
            """
            // RAR v4.x ends with 00, RAR v5.x ends with 01 00
            52 61 72 21 1A 07 ( 00 | 01 00 )
        """
        )
    ]
    EXTRACTOR = Command("unar", "-no-directory", "-p", "", "{inpath}", "-o", "{outdir}")

    DOC = HandlerDoc(
        name="RAR",
        description="RAR archive files are commonly used for compressed data storage. They can contain multiple files and directories, and support various compression methods.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="RAR 4.x File Format Documentation",
                url="https://codedread.github.io/bitjs/docs/unrar.html",
            ),
            Reference(
                title="RAR 5.x File Format Documentation",
                url="https://www.rarlab.com/technote.htm#rarsign",
            ),
        ],
        limitations=["Does not support encrypted RAR files."],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> ValidChunk | None:
        try:
            rar_file = rarfile.RarFile(file)
        except (rarfile.Error, ValueError):
            return None

        # RarFile has the side effect of moving the file pointer
        rar_end_offset = file.tell()

        return ValidChunk(
            start_offset=start_offset,
            end_offset=rar_end_offset,
            is_encrypted=rar_file.needs_password(),
        )
